# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
This file gives an API for causal inference
"""
from ._doorCriteria import *
from ._dSeparation import *
from ._doAST import *

from ._exceptions import *
from ._CausalModel import CausalModel
from ._CausalFormula import CausalFormula, _getLabelIdx
from ._doCalculus import doCalculusWithObservation, doCalculus, getFrontDoorTree, getBackDoorTree


def causalImpact(cm: CausalModel, on: Union[str, NameSet],
                 doing: Union[str, NameSet],
                 knowing: Optional[NameSet] = None,
                 values: Optional[Dict[str, int]] = None) -> Tuple[CausalFormula, gum.Potential, str]:
  """
  Determines the causal impact of interventions.

  Determines the causal impact of the interventions specified in ``doing`` on the single or list of variables ``on``
  knowing the states of the variables in ``knowing`` (optional). These last parameters is dictionary <variable
  name>:<value>. The causal impact is determined in the causal DAG ``cm``.
  This function returns a triplet with a latex format formula used to compute the causal impact, a potential
  representing the probability distribution of ``on``  given the interventions and observations as parameters,
  and an explanation of the method allowing the identification. If there is no impact, the joint probability of
  ``on`` is simply returned. If the impact is not identifiable the formula and the adjustment will be ``None`` but an
  explanation  is still given.

  :param cm: causal model
  :param on: variable name or variable names set
  :param doing: variable name or variable names set
  :param knowing: variable names set
  :param values: Dictionary
  :return: the CausalFormula, the computation, the explanation
  """
  # Checking the args
  son = {on} if isinstance(on, str) else on
  sdoing = {doing} if isinstance(doing, str) else doing
  sk = set() if knowing is None else knowing

  # verifies that when len()==0, args are empty sets
  if len(son) == 0:
    son = set()
  if len(sdoing) == 0:
    sdoing = set()
  if len(sk) == 0:
    sk = set()

  total = {cm.bn().variable(cm.bn().idFromName(i)).name() for i in son | sdoing | sk}

  if values is not None:
    for k in values.keys():
      if k not in total:
        raise (ValueError, f"{k} is not in the query arguments.")

  if len(son & sdoing & sk) > 0:
    raise (ValueError, f"The 3 parts of the query (on, doing, knowing) must not intersect.")

  formula, potential, explanation = _causalImpact(cm, son, sdoing, sk)

  # no need to contextualize the potential
  if potential is None or values is None:
    return formula, potential, explanation

  sv = set(potential.var_names)
  extract_values = {k: _getLabelIdx(cm.bn(), k, v) for k, v in values.items() if k in sv}
  potextract = potential.extract(extract_values)
  return formula, potextract, explanation


def _causalImpact(cm: CausalModel, on: NameSet,
                  doing: NameSet,
                  knowing: NameSet) -> Tuple[CausalFormula, gum.Potential, str]:
  """
  Determines the causal impact of interventions.

  Determines the causal impact of the interventions specified in ``doing`` on the single or list of variables ``on``
  knowing the states of the variables in ``knowing`` (optional). The causal impact is determined in the causal DAG
  ``cm``.
  This function returns a triplet with a latex format formula used to compute the causal impact, a potential
  representing the probability distribution of ``on``  given the interventions and observations as parameters,
  and an explanation of the method allowing the identification. If there is no impact, the joint probability of
  ``on`` is simply returned. If the impact is not identifiable the formula and the adjustment will be ``None`` but an
  explanation  is still given.

  :param cm: causal model
  :param on: variable name or variable names set
  :param doing: variable name or variable names set
  :param knowing: variable names set
  :return: the latex representation, the computation, the explanation
  """
  nY = [y for y in on]
  iY = [cm.bn().idFromName(i) for i in nY]

  nDo = [d for d in doing]
  iDo = [cm.bn().idFromName(i) for i in nDo]

  if knowing is None:
    nK = list()
  else:
    nK = [k for k in knowing]
  sK = {cm.bn().idFromName(i) for i in nK}

  # Null causal effect (different from non identifiable causal effect)
  if isDSep(cm, set(iDo), set(iY), sK | cm.latentVariablesIds()):
    explain = "No causal effect of X on Y, because they are d-separated "
    explain += "(conditioning on the observed variables if any)."
    ar = CausalFormula(cm, ASTposteriorProba(cm.bn(), set(nY), set(nK)), on, doing, knowing)
    adj = ar.eval()
    return ar, adj, explain

  # Front or Back door
  if len(iDo) == 1 and len(nY) == 1 and len(nK) == 0:
    for bd in backdoor_generator(cm, iDo[0], iY[0], cm.latentVariablesIds()):
      ar = CausalFormula(cm, getBackDoorTree(cm, nDo[0], nY[0], bd), on, doing, knowing)
      adj = ar.eval()
      explain = "backdoor " + str([cm.bn().variable(i).name() for i in bd]) + " found."
      return ar, adj, explain

    for fd in frontdoor_generator(cm, iDo[0], iY[0], cm.latentVariablesIds()):
      ar = CausalFormula(cm, getFrontDoorTree(cm, nDo[0], nY[0], fd), on, doing, knowing)
      adj = ar.eval()
      explain = "frontdoor " + str([cm.bn().variable(i).name() for i in fd]) + " found."
      return ar, adj, explain

  # Go for do-calculus
  try:
    if len(nK) == 0:
      ar = doCalculus(cm, on, set(nDo))
    else:
      ar = doCalculusWithObservation(cm, on, set(nDo), set(nK))
  except HedgeException as h:
    return None, None, h.message

  adj = ar.eval()
  explain = "Do-calculus computations"

  return ar, adj, explain



# A Python Library to create a Progress Bar.
# from http://snipperize.todayclose.com/snippet/py/Creating-a-Progress-Bar-in-Console--14251/
# modified by Pierre-Henri Wuillemin (feb 2011)
from __future__ import print_function
import sys

class ProgressBar:
    def __init__(self, title="",min_value = 0, max_value = 100, width=77,**kwargs):
        self.char = kwargs.get('char', '#')
        self.mode = kwargs.get('mode', 'dynamic') # fixed or dynamic
        if not self.mode in ['fixed', 'dynamic']:
            self.mode = 'fixed'

        self.total_width=width

        self.bar = ''
        self.oldvar = ''
        self.hasChanged = True

        self.min = min_value
        self.max = max_value
        self.span = max_value - min_value
        self.amount = 0       # When amount == max, we are 100% done

        self.change_title(title)
        self.update_amount(0)

    def change_title(self,new_title):
        self.title = new_title
        self.width=self.total_width-len(self.title)
        self.build_bar()

    def increment_amount(self, add_amount = 1):
        """
        Increment self.amount by 'add_ammount' or default to incrementing
        by 1, and then rebuild the bar string.
        """
        new_amount = self.amount + add_amount
        if new_amount < self.min: new_amount = self.min
        if new_amount > self.max: new_amount = self.max
        self.amount = new_amount
        self.build_bar()


    def update_amount(self, new_amount = None):
        """
        Update self.amount with 'new_amount', and then rebuild the bar
        string.
        """
        if not new_amount: new_amount = self.amount
        if new_amount < self.min: new_amount = self.min
        if new_amount > self.max: new_amount = self.max
        self.amount = new_amount
        self.build_bar()


    def build_bar(self):
        """
        Figure new percent complete, and rebuild the bar string base on
        self.amount.
        """
        self.oldbar=self.bar

        diff = float(self.amount - self.min)
        percent_done = int(round((diff / float(self.span)) * 100.0))

        # figure the proper number of 'character' make up the bar
        all_full = self.width - 2
        num_hashes = int(round((percent_done * all_full) / 100))

        if self.mode == 'dynamic':
            # build a progress bar with self.char (to create a dynamic bar
            # where the percent string moves along with the bar progress.
            self.bar = self.char * num_hashes
        else:
            # build a progress bar with self.char and spaces (to create a
            # fixe bar (the percent string doesn't move)
            self.bar = self.char * num_hashes + ' ' * (all_full-num_hashes)

        percent_str = str(percent_done) + "%"
        self.bar = self.title+'[ ' + self.bar + ' ] ' + percent_str
        self.hasChanged=(self.oldbar!=self.bar)


    def __str__(self):
        return str(self.bar)

    def display(self):
        """
        Display bar if needed
        """
        if self.hasChanged:
            print('\r', self.bar,end="")
            sys.stdout.flush()
            self.hasChanged=False



if __name__ == '__main__':
    print
    limit = 1000000

    print('Example 1: Fixed Bar')
    prog = ProgressBar("ex1",0, limit, 77, mode='fixed')
    oldprog = str(prog)
    for i in xrange(limit+1):
        prog.update_amount(i)
        if oldprog != str(prog):
            print("\r",prog,end="",flush=True)
            oldprog=str(prog)

    print('\n\n')

    print('Example 2: Dynamic Bar')
    prog = ProgressBar("ex2",0, limit, 77, mode='dynamic', char='-')
    oldprog = str(prog)
    for i in xrange(limit+1):
        prog.increment_amount()
        if oldprog != str(prog):
            print("\r",prog,end="",flush=True)
            oldprog=str(prog)

    print('\n\n')

    print('Example 3: Dynamic Bar')
    prog = ProgressBar("ex3",0, limit, 77, mode='dynamic', char='-')
    oldprog = str(prog)
    for i in xrange(limit+1):
        prog.increment_amount()
        prog.display()

    print('\n\n')

#<code>
#from progress_bar import ProgressBar
#import sys
#
#count = 0
#total = 100000
#
#prog = ProgressBar(count, total, 77, mode='fixed', char='#')
#while count <= total:
#    count += 1
#    prog.increment_amount()
#		 prog.display()
#print
#</code>



# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
This file defines functions for dSeparation crtieria
"""
from ._types import *

import pyAgrum as gum


def isParent(a: NodeId, b: NodeId, g: gum.BayesNet) -> bool:
  """
  Predicate on whether ``a`` is parent of ``b`` in the graph ``g``, the graph must be a DAG
  """
  return g.existsArc(a, b)


def ancester(x: NodeId, dm: DirectedModel, anc: NodeSet):
  """
  Adds the ancestors of ``x`` in the Bayesian network ``bn`` to the set ``P``
  """
  for parent in dm.parents(x):
    if parent not in anc:
      anc.add(parent)
      ancester(parent, dm, anc)


def _reduce_moralize(bn: gum.BayesNet, x: NodeSet, y: NodeSet, zset: NodeSet) -> gum.UndiGraph:
  """
  Returns the undirected graph obtained by reducing (ancestor graph) and moralizing the Bayesian network ``bn``

  :param bn: the BayesNet
  :param x: NodeSet generating the ancestor graph
  :param y: Second NodeSet generating the ancestor graph
  :param zset: Third NodeSet generating the ancestor graph
  :return:
  """
  G = gum.UndiGraph()

  Ancetre = x | y
  anc = frozenset(Ancetre)
  for i in anc:
    ancester(i, bn, Ancetre)

  for i in zset:
    Ancetre.add(i)
    ancester(i, bn, Ancetre)
  for i in Ancetre:
    G.addNodeWithId(i)

  for b in G.nodes():
    for a in bn.parents(b):
      G.addEdge(a, b)

  for nod in G.nodes():
    parents_nod = bn.parents(nod)
    for par in parents_nod:
      for par2 in parents_nod:
        if par2 != par:
          G.addEdge(par, par2)

  return G


def _removeZ(g_undi: gum.UndiGraph, zset: NodeSet):
  for node in g_undi.nodes():
    if node in zset:
      g_undi.eraseNode(node)


def _is_path_x_y(g_undi: gum.UndiGraph, sx: NodeSet, sy: NodeSet, marked: NodeSet = None) -> bool:
  """
  Predicate asserting the existence of a path between ``x`` and ``y`` in the non oriented graph
  ``g_undi``, without going through the optional marking set ``mark``

  :param g_undi:
  :param x:
  :param y:
  :param marked:
  :return: True if a path has been found
  """

  def inner_ec(g_und: gum.UndiGraph, a: NodeId, b: NodeSet, m: NodeSet) -> bool:
    if a in b:
      return True

    m.add(a)

    for n in g_und.neighbours(a):
      if n not in m:
        if inner_ec(g_und, n, b, m):
          return True

    return False

  if len(sx) > len(sy):
    ssx, ssy = sy, sx
  else:
    ssx, ssy = sx, sy

  if marked is None:
    marked = set()

  ma = set(marked)
  for i in ssx:
    ma.add(i)
    if inner_ec(g_undi, i, ssy, ma):
      return True

  return False


def isDSep_tech2(bn: gum.BayesNet, sx: NodeSet, sy: NodeSet, zset: NodeSet) -> bool:
  """
  Test of d-separation for ``x`` and ``y``, given ``zset`` using the graph-moralization method

  :param bn: the bayes net
  :param x: 
  :param y:
  :param zset:
  :return: True if ``Z`` d-separates ``x`` and ``y``
  """
  g_undi = _reduce_moralize(bn, sx, sy, zset)

  _removeZ(g_undi, zset)

  if _is_path_x_y(g_undi, sx, sy):
    return False

  return True


def isDSep_parents(bn: gum.BayesNet, sx: NodeSet, sy: NodeSet, zset: NodeSet) -> bool:
  """Test of d-separation of ``sx`` and ``sy`` given ``Z``, considering only the paths with an arc coming into ``x``
  using the graph-moralization method"""
  return _isDSep_tech2_parents(bn, sx, sy, zset)


def _isDSep_tech2_parents(bn: gum.BayesNet, sx: NodeSet, sy: NodeSet, zset: NodeSet) -> bool:
  """Test of d-separation of ``sx`` and ``sy`` given ``Z``, considering only the paths with an arc coming into ``x``
  using the graph-moralization method"""
  G = gum.UndiGraph()
  ancesters = sx | sy
  anc = frozenset(ancesters)
  for i in anc:
    ancester(i, bn, ancesters)

  for i in zset:
    ancesters.add(i)
    ancester(i, bn, ancesters)
  for i in ancesters:
    G.addNodeWithId(i)

  for b in G.nodes():
    for a in (set(bn.parents(b)) - sx):
      G.addEdge(a, b)

  for nod in G.nodes():
    parents_nod = set(bn.parents(nod)) - sx
    for par in parents_nod:
      for par2 in parents_nod:
        if par2 != par:
          G.addEdge(par, par2)

  _removeZ(G, zset)

  if _is_path_x_y(G, sx, sy):
    return False

  return True


def _isDSep_tech2_children(bn: gum.BayesNet, sx: NodeSet, sy: NodeSet, zset: NodeSet) -> bool:
  """Test of d-separation of ``x`` and ``y`` given ``zset``, considering only the paths with an arc coming from ``x``
  using the graph-moralization method"""
  G = gum.UndiGraph()
  ancesters = sx | sy
  for i in sy:
    ancester(i, bn, ancesters)
  # sx's ancesters will not be added since sx already is in ancesters
  for i in zset:
    ancesters.add(i)
    ancester(i, bn, ancesters)
  for i in ancesters:
    G.addNodeWithId(i)
  se = set(G.nodes()) - sx
  for b in se:
    for a in bn.parents(b):
      G.addEdge(a, b)

  for nod in se:
    parents_nod = bn.parents(nod)
    for par in parents_nod:
      for par2 in parents_nod:
        if par2 != par:
          G.addEdge(par, par2)
  _removeZ(G, zset)

  if _is_path_x_y(G, x, y):
    return False

  return True


def _is_descendant(bn: gum.BayesNet, x: NodeId, y: NodeId, marked: NodeSet = None) -> bool:
  """ Asserts whether or not ``x`` is a descendant of ``y`` in ``bn`` """

  if marked is None:
    marked = set()

  if isParent(y, x, bn):
    return True

  for c in bn.children(y):
    if c not in marked:
      marked.add(c)
      if _is_descendant(bn, x, c, marked):
        return True

  return False


def _is_ascendant(bn: gum.BayesNet, x: NodeId, y: NodeId, marquage: Set[int] = None) -> bool:
  """Predicate on whether ``x`` is an ancestor of ``y`` in the Bayesian network ``bn``"""

  if isParent(x, y, bn):
    return True

  if marquage is None:
    marquage = set()

  for p in bn.parents(y):
    if p not in marquage:
      marquage.add(p)
      if _is_ascendant(bn, x, p, marquage):
        return True

  return False


def descendants(bn: gum.BayesNet, x: NodeId, marked: NodeSet = None, ensdesc: NodeSet = None) -> NodeSet:
  """ Returns a set composed by all the descendents of ``x`` in ``bn`` """
  if marked is None:
    marked = set()
  if ensdesc is None:
    ensdesc = set()

  ensdesc = ensdesc | set(bn.children(x))

  for c in bn.children(x):
    if c not in marked:
      marked.add(c)
      ensdesc = ensdesc | descendants(bn, c, marked)

  return ensdesc


def _filaires(bn: DirectedModel, interest: NodeSet = None, inf: bool = True) -> NodeSet:
  s = set()

  if interest is None:
    interest = set()

  for x in bn.nodes():
    if len(set(bn.parents(x)) - s) == 0 and len(bn.children(x)) == 1 and x not in interest:
      a = x
      while True:
        s.add(a)
        for a in bn.children(a):
          break  # take the first elt
        if len(bn.children(a)) != 1 or len(set(bn.parents(a)) - s) != 0 or a in interest:
          break

    if inf and len(bn.children(x)) == 0 and len(bn.parents(x)) == 1 and x not in interest:
      a = x
      while True:
        s.add(a)
        for a in bn.parents(a):
          break  # take the first elt
        if len(bn.children(a)) != 1 or a in interest:
          break
        if len(bn.parents(a)) != 1:
          s.add(a)
          break
  return s


def _barren_nodes(bn: gum.BayesNet, interest: NodeSet = None) -> NodeSet:
  """Returns the set of recursively determined barren nodes in ``bn`` relatively to the set of nodes ``interest`` (if
  ``interest`` is void, then the whole set of nodes in the graph will be returned)"""
  s = set()

  if interest is None:
    interest = set()

  def inner_rec(a):
    if a in interest | s:
      return
    s.add(a)
    for b in bn.parents(a):
      if len(set(bn.children(b)) - s) == 0 and (b not in s):
        inner_rec(b)

  for x in bn.nodes():
    if len(bn.children(x)) == 0:
      inner_rec(x)

  return s


def partialDAGFromBN(bn: gum.BayesNet, Nexcl: NodeSet = None) -> gum.DAG:
  """ Creates and returns a duplicate DAG of the given Bayesian network """

  if Nexcl is None:
    Nexcl = set()
  d = gum.DAG()

  nodes = set(bn.nodes()) - (Nexcl)
  for n in nodes:
    d.addNodeWithId(n)

  for x, y in bn.arcs():
    if x in nodes and y in nodes:
      d.addArc(x, y)

  return d


def dSep_reduce(g: gum.BayesNet, interest: NodeSet = None) -> gum.DAG:
  if interest is None:
    interest = set()

  barren = _barren_nodes(g, interest)

  reduced_g = partialDAGFromBN(g, barren)

  for f in _filaires(reduced_g, interest, False):
    reduced_g.eraseNode(f)

  return reduced_g


def blocked(bn: gum.BayesNet, pht: bool, x: NodeSet, y: NodeSet, setz: NodeSet,
            marquage0: Set[int],
            marquage1: Set[int]) -> bool:
  if x in y:
    return False

  isInxZ = x in setz
  wasIn = x in marquage0 or x in marquage1

  if pht:
    marquage1.add(x)
  else:
    marquage0.add(x)

  if not isInxZ and not wasIn:
    for c in bn.children(x):
      if c not in marquage1 and not blocked(bn, True, c, y, setz, marquage0, marquage1):
        return False

  if pht:
    if isInxZ or len(setz & descendants(bn, x)) != 0:
      for p in bn.parents(x):
        if p not in marquage0 and not blocked(bn, False, p, y, setz, marquage0, marquage1):
          return False

  else:
    if not isInxZ:
      for p in bn.parents(x):
        if p not in marquage0 and not blocked(bn, False, p, y, setz, marquage0, marquage1):
          return False

  return True


def _isDSep_tech1_parents(bn: gum.BayesNet, x: NodeId, sy: NodeSet, zset: NodeSet, reduced: bool = False) -> bool:
  """ Test of d-separation of ``x`` and ``y`` given ``Z``, considering only the paths with an arc coming into ``x``
  using the usual paths method"""

  if not reduced and len(bn.nodes()) > 170:
    g = dSep_reduce(bn, sy | zset | {x})
  else:
    g = bn

  for p in g.parents(x):
    if not blocked(g, False, p, sy, zset, {x}, {x}):
      return False
  return True


def _isDSep_tech1_children(bn: gum.BayesNet, x: NodeId, sy: NodeSet, setz: NodeSet, reduced=False) -> bool:
  """ Test of d-separation of ``x`` and ``y`` given ``Z``, considering only the paths with an arc coming from ``x``
  using the usual paths method"""

  if not reduced and len(bn.nodes()) > 170:
    g = dSep_reduce(bn, sy | setz | {x})
  else:
    g = bn

  for c in g.children(x):
    if not blocked(g, True, c, sy, setz, {x}, {x}):
      return False
  return True


def isDSep_tech1(bn: gum.BayesNet, x: NodeSet, y: NodeSet, setz: NodeSet, reduced=False) -> bool:
  """ Test of d-separation for ``x`` and ``y``, given ``Z`` using the usual paths method """
  sy = setify(y)
  sx = setify(x)

  if len(sx) > len(sy):
    sx, sy = sy, sx

  if not reduced and len(bn.nodes()) > 170:
    g = dSep_reduce(bn, sx | sy | setz)
  else:
    g = bn
  for i in sx:
    if not _isDSep_tech1_parents(g, i, sy, setz, True):
      return False
    if not _isDSep_tech1_children(g, i, sy, setz, True):
      return False
  return True


def isDSep(bn: gum.BayesNet, x: NodeSet, y: NodeSet, setz: NodeSet) -> bool:
  return isDSep_tech2(bn, x, y, setz)



# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
This file computes the causal impact of intervention in a causal model
"""

from ._doorCriteria import *
from ._dSeparation import *
from ._doAST import *

from ._exceptions import *
from ._CausalModel import inducedCausalSubModel, CausalModel
from ._CausalFormula import CausalFormula


def doCalculusWithObservation(cm: CausalModel, on: str, doing: NameSet, knowing: Optional[NameSet] = None) -> CausalFormula:
  """
  Compute the CausalFormula for an impact analysis given the causal model, the observed variables and the
  variable on  which there will be intervention.

  :param on: the variables of interest
  :param cm: the causal model
  :param doing: the interventions
  :param knowing: the observations
  :return: the CausalFormula for computing this causal impact
  """
  if knowing is None or len(knowing) == 0:
    return doCalculus(cm, on, doing)

  lOn = [i for i in on]
  iDoing = {cm.idFromName(i) for i in doing}
  iOn = {cm.idFromName(i) for i in lOn}
  iKnowing = {cm.idFromName(i) for i in knowing}

  removedArcs = set()

  for e in iDoing:
    for p in cm.parents(e):
      cm.eraseCausalArc(p, e)
      removedArcs.add((p, e))
  for e in iKnowing:
    for c in cm.children(e):
      cm.eraseCausalArc(e, c)
      removedArcs.add((e, c))
  rg = dSep_reduce(cm, iDoing | iOn | iKnowing)
  for a, b in removedArcs:
    cm.addCausalArc(a, b)

  for i in iKnowing:
    if isDSep(rg, {i}, iOn, iDoing | (iKnowing - {i})):
      try:
        return doCalculusWithObservation(cm, on, doing | {cm.names()[i]}, knowing - {cm.names()[i]})
      except HedgeException as h:
        pass

  p = doCalculus(cm, lOn | Z, X)
  q = doCalculus(cm, Z, X)

  return CausalFormula(cm, ASTdiv(p.root, q.root))


def _cDecomposition(cm: CausalModel) -> List[Set[int]]:
  undi = gum.UndiGraph()
  s = set(cm.nodes()) - cm.latentVariablesIds()
  for n in s:
    undi.addNodeWithId(n)
  for a, b in cm.biArcs():
    undi.addEdge(a, b)

  def undiCComponent(g, n, se):
    for i in g.neighbours(n):
      if i not in se:
        se.add(i)
        undiCComponent(g, i, se)

  components = []
  while len(s) != 0:
    c = s.pop()
    sc = set([c])
    undiCComponent(undi, c, sc)
    s -= sc
    components.append(sc)
  return components


def _topological_sort(cm: CausalModel) -> List[int]:
  lt = cm.latentVariablesIds()

  dc = {i: len(set(cm.parents(i)) - lt) for i in (set(cm.nodes()) - lt)}

  order = []
  while len(dc) != 0:
    rem = set()
    for i in dc:
      if dc[i] == 0:
        order.append(i)
        rem.add(i)

    for r in rem:

      for c in cm.children(r):
        dc[c] -= 1
      del dc[r]

  return order


def doCalculus(cm: CausalModel, on: Union[str, NameSet], doing: Union[str, NameSet]) -> CausalFormula:
  """
  Compute the CausalFormula for computing an impact analysis given the causal model, the observed variables and the
  variable on
  which there will be intervention. Note that there is no value neither for ``on`` nor for ``doing`` variables

  :param on: the variables of interest
  :param cm: the causal model
  :param doing: the interventions
  :return: the CausalFormula for computing this causal impact
  """
  X = doing if type(doing) == set else {doing}  # set of keys in doing
  Y = on if type(on) == set else {on}

  return CausalFormula(cm, identifyingIntervention(cm, Y, X), on, doing)


def identifyingIntervention(cm: CausalModel, Y: NameSet, X: NameSet, P: ASTtree = None) -> ASTtree:
  """
  Following Shpitser, Ilya and Judea Pearl. 'Identification of Conditional Interventional Distributions.' UAI2006 and
  'Complete Identification Methods for the Causal Hierarchy' JMLR 2008

  :param cm: the causal model
  :param Y: The variables of interest (named following the paper)
  :param X: The variable of intervention (named following the paper)
  :param P: The ASTtree representing the calculus in construction
  :return: the ASTtree representing the calculus
  """
  iX = {cm.idFromName(i) for i in X}
  iY = {cm.idFromName(i) for i in Y}
  iV = set(cm.nodes()) - cm.latentVariablesIds()
  V = {cm.names()[i] for i in iV}

  # 1 ------------------------------------------------
  if len(X) == 0:
    if P is None:
      lY = list(Y)
      return ASTjointProba(lY)
    else:
      vy = V - Y
      if len(vy) != 0:
        lvy = list(vy)
        ilvy = [cm.idFromName(i) for i in lvy]
        return ASTsum(lvy, ilvy, P)
      return P

  # 2 -------------------------------------------
  iAnY = set()
  for i in iY:
    ancester(i, cm, iAnY)
  iAnY |= iY
  AnY = {cm.names()[i] for i in iAnY}
  iLat = cm.latentVariablesIds()
  Lat = {cm.names()[i] for i in iLat}

  if len(cm.nodes()) != len(AnY):
    ivAny = list(iV - iAnY)

    if P is not None:
      vAny = [cm.names()[j] for j in ivAny]
      P = ASTsum(vAny, P)
    return identifyingIntervention(inducedCausalSubModel(cm, iAnY), Y, X & AnY, P)

  # 3  -----------------------------------------------
  rmArcs = set()

  for e in iX:
    for p in cm.parents(e):
      cm.eraseCausalArc(p, e)
      rmArcs.add((p, e))

  ianY = set()
  for i in iY:
    ancester(i, cm, ianY)
  ianY |= iY
  iW = (iV - iX) - ianY

  for a, b in rmArcs:
    cm.addCausalArc(a, b)

  if len(iW) != 0:
    W = {cm.names()[i] for i in iW}
    return identifyingIntervention(cm, Y, X | W, P)

  gvx = inducedCausalSubModel(cm, iV - iX)
  icd = _cDecomposition(gvx)
  cd = []
  for s in icd:
    cd.append({cm.names()[i] for i in s})

  # 4----------------
  if len(cd) > 1:
    t = identifyingIntervention(cm, cd[0], V - cd[0], P)
    for si in cd[1:]:
      Pp = copyAST(P) if P is not None else None
      t = ASTmult(identifyingIntervention(cm, si, V - si, Pp), t)

    vyx = V - (X | Y)
    lvyx = list(vyx)

    if len(lvyx) == 0:
      return t
    return ASTsum(lvyx, t)

  else:
    S = cd[0]
    iS = icd[0]
    cdg = _cDecomposition(cm)

    # 5-------------------------
    if len(cdg) == 1 and len(cdg[0]) == len(V):
      raise HedgeException("Hedge Error: " + "G=" + str(V) + ",G[S]=" + str(S), V, S)

    # 6--------------------------
    gs = inducedCausalSubModel(cm, iS)
    if (set(gs.nodes()) - gs.latentVariablesIds()) in cdg:
      vpi = []
      prb = []
      to = _topological_sort(cm)
      for v in S:
        vpi = to[:to.index(cm.idFromName(v))]
        nvpi = {cm.names()[i] for i in vpi}

        if len(nvpi) == 0:
          prb.append(ASTjointProba([v]))
        else:
          if P is None:
            prb.append(ASTposteriorProba(cm.bn(), {v}, nvpi))
          else:
            prb.append(ASTdiv(P.copy(), ASTsum(v, P.copy())))

      prod = productOfTrees(prb)

      if len(S - Y) == 0:
        return prod

      else:
        lsy = list(S - Y)
        ilsy = [cm.idFromName(i) for i in lsy]
        return ASTsum(lsy, ilsy, prod)

    # 7------------------------------------------
    for ispr in cdg:
      if iS <= ispr:
        spr = {cm.names()[i] for i in ispr}
        prb = []
        top = _topological_sort(cm)

        for v in spr:
          vpi = top[:top.index(cm.idFromName(v))]
          nvpi = {cm.names()[i] for i in vpi}

          if len(nvpi) == 0:
            prb.append(ASTjointProba([v]))
          else:
            prb.append(ASTposteriorProba(cm.bn(), {v}, nvpi))

        P = productOfTrees(prb)
        return identifyingIntervention(inducedCausalSubModel(cm, ispr), Y, X & spr, P)


def getBackDoorTree(cm: CausalModel, x: str, y: str, zset: NodeSet) -> ASTtree:
  """
  Create an CausalFormula representing a backdoor zset from x to y in the causal mode lcm

  :param cm: causal model
  :param x: impacting node
  :param y: impacted node
  :param zset: backdoor set
  :return: the ASTtree for the backoor criteria
  """
  zp = [cm.names()[i] for i in zset]
  return ASTsum(zp,
                ASTmult(ASTposteriorProba(cm.bn(), {y}, set([x] + zp)),
                        ASTjointProba(zp))
                )


def getFrontDoorTree(cm: CausalModel, x: str, y: str, zset: NodeSet) -> ASTtree:
  """
  Create an AdsT representing a frontdoor zset from x to y in the causal model

  :param cm: causal model
  :param x: impacting node
  :param y: impacted node
  :param zset: frontdoor set
  :return: the ASTtree for the frontdoot critreroia
  """
  zp = [cm.names()[i] for i in zset]
  return ASTsum(zp, ASTmult(ASTposteriorProba(cm.bn(), set(zp), {x}),
                            ASTsum([x],
                                   ASTmult(ASTposteriorProba(cm.bn(), {y}, set([x] + zp)),
                                           ASTjointProba([x]))
                                   )
                            )
                )



# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017
# (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

from __future__ import print_function

import sys
import os
import math

import pyAgrum as gum
import pydotplus as dot
from itertools import product

from ._utils.pyAgrum_header import pyAgrum_header


class BNComparator:
  """
  BNComparator allows to compare in multiple way 2 BNs...The smallest assumption is that the names of the variables
  are the same in the 2 BNs.
  But some comparisons will have also to check the type and domainSize of the variables. The bns have not exactly the
  same role :
  _bn1 is rather the referent model for the comparison whereas _bn2 is the compared one to the referent model

  Parameters
  ----------
  name1 : str or pyAgrum.BayesNet
  	a BN or a filename for reference
  name2 : str or pyAgrum.BayesNet
  	another BN or antoher filename for comparison
  """

  def __init__(self, name1, name2, delta=1e-6):
    self.DELTA_ERROR = delta
    if isinstance(name1, str):
      self._bn1 = gum.loadBN(name1)
      self._bn1.setProperty('name', '"' + os.path.basename(self._bn1.property('name') + '"'))
    else:
      self._bn1 = name1

    if isinstance(name2, str):
      self._bn2 = gum.loadBN(name2)
      self._bn2.setProperty('name', '"' + os.path.basename(self._bn2.property('name') + '"'))
    else:
      self._bn2 = name2

    s1 = set(self._bn1.names())
    s2 = set(self._bn2.names())

    if s1 != s2:
      raise ValueError(
          "The 2 BNs are not comparable! There are names not present in the 2 BNs : " + str(
              s1.symmetric_difference(s2)))

  def _compareBNVariables(self):
    """
    Returns
    -------
    str
      'OK' if BN are composed of the same variables, indicates problematic variables otherwise
  
    """
    # it is assumed (checked by the constructor) that _bn1 and _bn2 share the same set of variable names
    for i in self._bn1.nodes():
      v1 = self._bn1.variable(i)
      v2 = self._bn2.variableFromName(v1.name())
      if v2.domainSize() != v1.domainSize():
        return v1.name() + " has not the same domain size in the two bns"

    return "OK"

  def _compareBNParents(self):
    """
    Returns
    -------
    str
      'OK' if _bn2 have (at least) the same variable as b1 and their parents are the same.

    """
    for id1 in self._bn1.nodes():
      id2 = self._bn2.idFromName(self._bn1.variable(id1).name())

      p1 = _parents_name(self._bn1, id1)
      p2 = _parents_name(self._bn2, id2)
      if p1 != p2:
        return self._bn1.variable(id1).name() + " has different parents in the two bns whose names are in " + str(
            p1.symmetric_difference(p2))

    return "OK"

  def _comparePotentials(self, pot1, pot2):

    """
    Parameters
    ----------
    _bn1 : pyAgrum.BayesNet
      a Bayesian network
    pot1 : pyAgrum.Potential
      one of b1's cpts
    _bn2 : pyAgrum.BayesNet
      another Bayesian network
    pot2 : pyAgrum.Potential
      one of _bn2's cpts

    Returns
    -------
    str
      'OK' if CPTs are the same

    Raises
    ------
    gum.KeyError
      If cpts are not from the same variable
    """
    I1 = gum.Instantiation(pot1)
    I2 = gum.Instantiation(pot2)
    I1.setFirst()
    while not I1.end():
      I2.fromdict(I1.todict())  # copy value on the base of names
      if abs(pot1.get(I1) - pot2.get(I2)) > self.DELTA_ERROR:
        return "Different CPTs for " + pot1.variable(0).name()
      I1 += 1
    return "OK"

  def _compareBNCPT(self):
    """
    Returns
    -------
    str
      'OK' if _bn2 have (at least) the same variable as b1 and their cpts are the same
    """
    for i in self._bn1.nodes():
      res = self._comparePotentials(self._bn1.cpt(i), self._bn2.cpt(self._bn1.variable(i).name()))
      if res != "OK":
        return res
    return "OK"

  def equivalentBNs(self):
    """
    Parameters
    ----------
    name1 : str or pyAgrum.BayesNet
      a BN or a filename
    name2 : str or pyAgrum.BayesNet
      another BN or antoher filename

    Returns
    -------
    str
      "OK" if bn are the same, a description of the error otherwise

    """

    ret = self._compareBNVariables()
    if ret != "OK":
      return ret

    ret = self._compareBNParents()
    if ret != "OK":
      return ret

    ret = self._compareBNCPT()
    return ret

  def dotDiff(self):
    """ Return a pydotplus graph that compares the arcs of _bn1 (reference) with those of self._bn2.
    full black line: the arc is common for both
    full red line: the arc is common but inverted in _bn2
    dotted black line: the arc is added in _bn2
    dotted red line: the arc is removed in _bn2
        
    @author : Mélanie Munch

    Returns
    -------
    pydotplus.Dot
      the result dot graph
    """
    graph = dot.Dot(graph_type='digraph', bgcolor="transparent")

    # Nodes
    for n in self._bn1.names():
      node = dot.Node('"' + n + '"', style="filled",
                      bgcol="#444444",
                      fgcol="#FFFFFF",
                      tooltip='"({0}) {1}{2}"'.format(self._bn1.idFromName(n), n, ""))
      graph.add_node(node)

    # Arcs
    for n1 in self._bn2.names():
      for n2 in self._bn2.names():
        id1in2 = self._bn2.idFromName(n1)
        id2in2 = self._bn2.idFromName(n2)
        id1in1 = self._bn1.idFromName(n1)
        id2in1 = self._bn1.idFromName(n2)

        if self._bn1.dag().existsArc(id1in1, id2in1):
          if self._bn2.dag().existsArc(id1in2, id2in2):
            # If present in both
            edge = dot.Edge('"' + n1 + '"', '"' + n2 + '"')
            graph.add_edge(edge)
          elif self._bn2.dag().existsArc(id2in2, id1in2):
            # If inversed in _bn2
            edge = dot.Edge('"' + n2 + '"', '"' + n1 + '"')
            edge.set_color("red")
            graph.add_edge(edge)
          else:
            # If removed from _bn2
            edge = dot.Edge('"' + n1 + '"', '"' + n2 + '"', style='dashed')
            edge.set_color("red")
            graph.add_edge(edge)
        else:
          if not self._bn1.dag().existsArc(id2in1, id1in1) and self._bn2.dag().existsArc(id1in2, id2in2):
            # If added to _bn2
            edge = dot.Edge('"' + n1 + '"', '"' + n2 + '"', style='dashed')
            graph.add_edge(edge)

    return graph

  def scores(self):
    """
    Compute Precision, Recall, F-score for self._bn2 compared to self._bn1

    precision and recall are computed considering BN1 as the reference

    $Fscore=\frac{2\cdot recall\cdot precision}{recall+precision}$ and is the weighted average of Precision and Recall.

    $dist2opt=\sqrt{(1-precision)^2+(1-recall)^2}$ and represents the euclidian distance to the ideal(precision=1,
    recall=1)

    @author : Mélanie Munch

    Returns
    -------
      dict[str,double]
      A dictionnary containing 'precision', 'recall', 'fscore', 'dist2opt' and so on.
    """
    # t: True, f: False, p: Positive, n: Negative
    count = {"tp": 0, "tn": 0, "fp": 0, "fn": 0}

    # We look at all combination
    listVariables = self._bn1.names()
    for head, tail in product(listVariables, listVariables):
      if head != tail:
        idHead_1 = self._bn1.idFromName(head)
        idTail_1 = self._bn1.idFromName(tail)

        idHead_2 = self._bn2.idFromName(head)
        idTail_2 = self._bn2.idFromName(tail)

        if self._bn1.dag().existsArc(idHead_1, idTail_1):  # Check arcs head->tail
          if self._bn2.dag().existsArc(idHead_2, idTail_2):  # if arc: Good Direction
            count["tp"] += 1
          elif self._bn2.dag().existsArc(idTail_2, idHead_2):  # If arc: Bad Direction
            count["fn"] += 1
          else:  # If no arc:
            count["fn"] += 1
        elif self._bn1.dag().existsArc(idTail_1, idHead_1):  # Check arcs n2->n1
          if self._bn2.dag().existsArc(idHead_2, idTail_2):  # if arc: Bad Direction
            count["fn"] += 1
          elif self._bn2.dag().existsArc(idTail_2, idHead_2):  # If arc: Good Direction
            count["tp"] += 1
          else:  # If no arc:
            count["fn"] += 1
        else:  # Check if no arc
          if self._bn2.dag().existsArc(idHead_2, idTail_2) or self._bn2.dag().existsArc(idTail_2, idHead_2):  # If arc
            count["fp"] += 1
          else:  # If no arc
            count["tn"] += 1

    # Compute the scores
    if count["tp"] + count["fn"] != 0:
      recall = (1.0 * count["tp"]) / (count["tp"] + count["fn"])
    else:
      recall = 0.0

    if count["tp"] + count["fp"] != 0:
      precision = (1.0 * count["tp"]) / (count["tp"] + count["fp"])
    else:
      precision = 0.0

    if precision + recall != 0.0:
      Fscore = (2 * recall * precision) / (recall + precision)
    else:
      Fscore = 0.0

    return {
      'count'    : count,
      'recall'   : recall,
      'precision': precision,
      'fscore'   : Fscore,
      'dist2opt' : math.sqrt((1 - precision) ** 2 + (1 - recall) ** 2)
    }


def _nodeId(bn, n):
  if type(n) == str:
    return bn.idFromName(n)
  else:
    return n


def _parents_name(bn, n):
  return {bn.variable(n).name() for n in bn.parents(_nodeId(bn, n))}


def module_help(exit_value=1):
  """
  defines help viewed if args are not OK on command line, and exit with exit_value
  """
  print(os.path.basename(sys.argv[0]),
        "_bn1.{" + gum.availableBNExts() + "} self._bn2.{" + gum.availableBNExts() + "}")
  sys.exit(exit_value)


if __name__ == "__main__":
  pyAgrum_header("2011-19")

  if len(sys.argv) != 3:
    module_help()

  cmp = BNComparator(sys.argv[1], sys.argv[2])
  print(cmp.equivalentBNs())



# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
This file defines a representation for causal model
"""
from __future__ import annotations

import itertools as it

from ._types import *


class CausalModel:
  """
  From an observational BNs and the description of latent variables, this class represent a complet causal model
  obtained by adding the latent variables specified in ``latentVarsDescriptor`` to the Bayesian network ``bn``.

  :param bn: a observational bayesian network
  :param latentVarsDescriptor:  list of couples (<latent variable name>, <list of affected variables' ids>).
  :param keepArcs: By default, the arcs between variables affected by a common latent variable will be removed but
         this can be avoided by setting ``keepArcs`` to ``True``
  """

  def __init__(self, bn: gum.BayesNet,
               latentVarsDescriptor: LatentDescriptorList = None,
               keepArcs: bool = False):
    self.__observationalBN = bn

    if latentVarsDescriptor is None:
      latentVarsDescriptor = []

    firstIdForLatentVariable = max(bn.nodes()) + 1

    # we have to redefine those attributes since the __observationalBN may be augmented by latent variables
    self.__dag = gum.DAG()
    self.__names: Dict[NodeId, str] = {}
    self.__ids: Dict[str, NodeId] = {}

    # nodes of BN
    for n in bn.nodes():
      self.__dag.addNodeWithId(n)
      self.__names.update({n: self.bn().variable(n).name()})
      self.__ids.update({self.bn().variable(n).name(): n})

    # arcs on BN
    for x, y in bn.arcs():
      if keepArcs or (not _sameLatency(bn.variable(x).name(),
                                       bn.variable(y).name(),
                                       latentVarsDescriptor)):
        self.__dag.addArc(x, y)

    # latent variables and arcs from latent variables
    self.__lat: NodeSet = set()
    for i in range(len(latentVarsDescriptor)):
      n, ls = latentVarsDescriptor[i]
      self.__dag.addNodeWithId(firstIdForLatentVariable + i)
      self.__names.update({i + firstIdForLatentVariable: n})
      self.__ids.update({n: +firstIdForLatentVariable})
      self.__lat.add(firstIdForLatentVariable + i)

      for item in ls:
        j = bn.idFromName(item) if isinstance(item, str) else item
        self.__dag.addArc(i + firstIdForLatentVariable, j)

    self.__biArcs = {i for lt in self.__lat for i in it.combinations(self.children(lt), 2)}

  def bn(self) -> gum.BayesNet:
    """
    :return: the observational Bayesian network
    """
    return self.__observationalBN

  def parents(self, x: NodeId) -> NodeSet:
    """
    From a NodeId, returns its parent (as a set of NodeId)

    :param x: the node
    :return:
    """
    return self.__dag.parents(x)

  def children(self, x: NodeId) -> NodeSet:
    """
    :param x: the node
    :return: 
    """
    return self.__dag.children(x)

  def names(self) -> Dict[NodeId, str]:
    """
    :return: the map NodeId,Name
    """
    return self.__names

  def idFromName(self, name: str) -> NodeId:
    """
    :param name: the name of the variable
    :return: the id of the variable
    """
    return self.__ids[name]

  def biArcs(self) -> LatentDescriptorList:
    """
    :return: list of descriptor of latent variables
    """
    return self.__biArcs

  def latentVariablesIds(self) -> NodeSet:
    """
    :return: the set of ids of latent variables in the causal model
    """
    return self.__lat

  def eraseCausalArc(self, x, y):
    self.__dag.eraseArc(x, y)

  def addCausalArc(self, x, y):
    self.__dag.addArc(x, y)

  def existsArc(self, a, b) -> bool:
    return self.__dag.existsArc(a, b)

  def nodes(self):
    return self.__dag.nodes()

  def arcs(self) -> ArcSet:
    return self.__dag.arcs()


def inducedCausalSubModel(cm: CausalModel, sns: NodeSet = None) -> CausalModel:
  if sns is None:
    sns = cm.nodes()
  nodes = sns - cm.latentVariablesIds()

  bn = gum.BayesNet()

  for n in nodes:
    bn.add(cm.bn().variable(n), n)

  for x, y in cm.arcs():
    if y in nodes:
      if x in nodes:
        bn.addArc(x, y)

  names = cm.names()
  latentVarsDescriptor = list()
  for latentVar in cm.latentVariablesIds():
    if cm.children(latentVar).issubset(nodes):
      latentVarsDescriptor.append((names[latentVar],
                                   list(cm.children(latentVar))))

  return CausalModel(bn, latentVarsDescriptor, True)


def _sameLatency(xxx: NodeId, yyy: NodeId, latent_set: LatentDescriptorList) -> bool:
  for _, ll in latent_set:
    if xxx in ll:
      if yyy in ll:
        return True
  return False



# !/usr/bin/python
# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017  (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

from __future__ import print_function
import sys
import os

from ._utils.pyAgrum_header import pyAgrum_header
import pyAgrum as gum

DECIMAL_LENGTH = 4
BLANK = ' '


def module_help(exit_value=1):
    """
    defines help viewed if args are not OK on command line, and exit with exit_value
    """
    print(os.path.basename(sys.argv[0]), "src.{" + gum.availableBNExts() + "}")
    sys.exit(exit_value)


def max_length(v):
    m = len(v.name())
    for i in range(v.domainSize()):
        m = max(m, len(v.label(i)))
    return m


def cpt2txt(cpt,digits=DECIMAL_LENGTH):
    """
    string representation of a gum.Potential

    :param cpt: the Potential to represent
    :return: the string representation
    """
    number_format = '{0:.' + str(digits) + 'f}'
    res = []

    size = (1 + 2 + digits) * cpt.variable(0).domainSize() - 1

    width = {}
    total_width = 0
    for j in range(1, cpt.nbrDim()):
        width[j] = 2 + max_length(cpt.variable(j))
        total_width += width[j] + 1
    total_width -= 1

    res.append(' ' * total_width + '|-' + '-' * size + '-|')

    line = '-' * total_width + '|'
    line += "{0:^{1}}".format(cpt.variable(0).name()[0:size + 2], size + 2)
    line += '|'
    res.append(line)

    line = ''
    for j in range(1, cpt.nbrDim()):
        line += "{0:^{1}}|".format(cpt.variable(j).name(), width[j])
    if line == '':
        line = '|'
    line += BLANK
    for j in range(cpt.variable(0).domainSize()):
        line += ("{0:^{1}}" + BLANK).format(cpt.variable(0).label(j)[0:digits + 2], digits + 2)
    line += '|'
    res.append(line)
    res.append(('-' * total_width) + '|' + '-' * (2 + size) + '|')

    i = gum.Instantiation(cpt)
    i.setFirst()
    while not (i.end()):
        line = ''
        for j in range(1, cpt.nbrDim()):
            line += "{0:^{1}}|".format(cpt.variable(j).label(i.val(j)), width[j])
        if line == '':
            line = '|'
        line += ' '
        for j in range(cpt.variable(0).domainSize()):
            line += number_format.format(cpt.get(i)) + ' '
            i.inc()
        line += '|'
        res.append(line)

    res.append(('-' * total_width) + '|' + '-' * (2 + size) + '|')
    return "\n".join(res)


def bn2txt(aBN):
    """
    Representation of all CPTs of a gum.BayesNet

    :param aBN: the bayes net or the name of the file
    :return:
    """
    if isinstance(aBN, str):
        bn = gum.loadBN(aBN)
    else:
        bn = aBN

    return "\n".join([cpt2txt(bn.cpt(nod)) for nod in bn.topologicalOrder()])


if __name__ == "__main__":
    pyAgrum_header(2011)

    if len(sys.argv) < 2:
        module_help()

    bn2txt(sys.argv[1])



# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
This file defines some helpers for handling causal concepts in notebooks
"""
from ._types import *

import IPython
import pyAgrum.lib.notebook as gnb
import pydotplus as dot

from ._CausalModel import CausalModel
from ._causalImpact import causalImpact


def getCausalModel(cm: CausalModel, size: str = "4") -> str:
  """
  return a HTML representing the causal model
  :param cm: the causal model
  :param size: passd
  :param vals:
  :return:
  """
  graph = dot.Dot(graph_type='digraph')
  for n in cm.nodes():
    if n not in cm.latentVariablesIds():
      bgcol = "#444444"
      fgcol = "#FFFFFF"
      st = "filled"
    else:
      bgcol = "#FF0000"
      fgcol = "#000000"
      st = "dashed"

    graph.add_node(dot.Node(cm.names()[n],
                            style=st,
                            fillcolor=bgcol,
                            fontcolor=fgcol))

  for a, b in cm.arcs():
    graph.add_edge(dot.Edge(cm.names()[a], cm.names()[b],
                            style="dashed" if a in cm.latentVariablesIds() else "filled"))

  graph.set_size(size)

  return IPython.display.SVG(graph.create_svg()).data


def showCausalModel(cm: CausalModel, size: str = "4"):
  """
  Shows a graphviz svg representation of the causal DAG ``d``
  """
  html = getCausalModel(cm, size)
  IPython.display.display(IPython.display.HTML("<div align='center'>" + html + "</div>"))


def getCausalImpact(model: CausalModel, on: Union[str, NameSet], doing: Union[str, NameSet],
                    knowing: Optional[NameSet] = None, values: Optional[Dict[str, int]] = None) -> Tuple[
  str, gum.Potential, str]:
  """
  return a HTML representing of the three values defining a causal impact : formula, value, explanation
  :param model: the causal model
  :param on: the impacted variable(s)
  :param doing: the variable(s) of intervention
  :param knowing: the variable(s) of evidence
  :param values : values for certain variables

  :return: a triplet (CausalFormula, gum.Potential, explanation)
  """
  formula, impact, explanation = causalImpact(model, on, doing, knowing, values)
  return gnb.getSideBySide(getCausalModel(model),
                           "?" if formula is None else (
                               '$$\\begin{equation}' + formula.toLatex() + '\\end{equation}$$'),
                           "No result" if formula is None else impact,
                           captions=["Causal Model", "Explanation : " + explanation,
                                     "Impact : $" + ("?" if formula is None else formula.latexQuery(values)) + "$"])


def showCausalImpact(model: CausalModel, on: Union[str, NameSet], doing: Union[str, NameSet],
                     knowing: Optional[NameSet] = None, values: Optional[Dict[str, int]] = None):
  """
  display a HTML representing of the three values defining a causal impact :  formula, value, explanation
  :param model: the causal model
  :param on: the impacted variable(s)
  :param doing: the variable(s) of intervention
  :param knowing: the variable(s) of evidence
  :param values : values for certain variables
  """
  html = getCausalImpact(model, on, doing, knowing, values)
  IPython.display.display(IPython.display.HTML(html))



# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
pyAgrum.lib is a set of python tools for pyAgrum.
"""
__author__      = "Pierre-Henri Wuillemin"
__copyright__   = "(c) 2016-2019 PARIS"



# -*- coding: utf-8 -*-
#(c) Copyright by Pierre-Henri Wuillemin, UPMC, 2011  (pierre-henri.wuillemin@lip6.fr)

#Permission to use, copy, modify, and distribute this
#software and its documentation for any purpose and
#without fee or royalty is hereby granted, provided
#that the above copyright notice appear in all copies
#and that both that copyright notice and this permission
#notice appear in supporting documentation or portions
#thereof, including modifications, that you make.

#THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
#WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
#SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
#OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
#IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
#ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
#OR PERFORMANCE OF THIS SOFTWARE!

"""
This file defines a represenatation of a causal query in a causal model
"""

from __future__ import annotations


from collections import defaultdict

from ._types import *
from ._doAST import ASTtree


class CausalFormula:
  """
  Represents a causal query in a causal model. The query is encoded as an CausalFormula that can be evaluated in the
  causal model : $P(on|knowing,\overhook (doing))$

  :param cm: the causal model
  :param root: the syntax tree as the root ASTtree
  :param on: the variable or the set of variables of interest
  :param doing: the intervention variables
  :param knowing : the observation variables
  """

  def __init__(self, cm: "CausalModel", root: ASTtree, on: Union(str, NameSet), doing: Union(str, NameSet),
               knowing: Optional[NameSet] = None):
    self._cm = cm
    self._root = root
    if isinstance(on, str):
      self._on = {on}
    else:
      self._on = on
    if isinstance(doing, str):
      self._doing = {doing}
    else:
      self._doing = doing
    if knowing is None:
      self._knowing = set()
    else:
      self._knowing = knowing

  def __str__(self, prefix: str = "") -> str:
    """
    :param prefix: a prefix for each line of the string representation
    :return:  the string version of the CausalFormula
    """
    return self.root.__str__(prefix)

  def latexQuery(self, values: Optional[Dict[str, str]] = None) -> str:
    """
    Returns a string representing the query compiled by this Formula. If values, the query is annotated with the
    values in the dictionary.

    :param values: the values to add in the query representation
    :return: the string representing the causal query for this CausalFormula
    """
    if values is None:
      values = dict()

    def _getVarRepresentation(v: str) -> str:
      if v not in values:
        return v
      else:
        bn = self.cm.bn()
        label = bn.variable(self.cm.idFromName(v)).label(_getLabelIdx(bn, v, values[v]))
        return v + "=" + label

    # adding values when necessary
    on = [_getVarRepresentation(k) for k in self._on]
    doing = [_getVarRepresentation(k) for k in self._doing]
    knowing = [_getVarRepresentation(k) for k in self._knowing]

    latexOn = ",".join(on)

    latexDo = ""
    if len(doing) > 0:
      latexDo = "\\hookrightarrow "
      latexDo += ", \\hookrightarrow ".join(doing)

    latexKnw = ""
    if len(knowing) > 0:
      if latexDo != "":
        latexKnw = ", "
      latexKnw += ",".join(knowing)

    return "P( " + latexOn + " \\mid " + latexDo + latexKnw + ")"

  def toLatex(self) -> str:
    """
    :return: a LaTeX representation of the CausalFormula
    """

    occur = defaultdict(int)
    for n in self._cm.bn().nodes():
      occur[self._cm.bn().variable(n).name()] = 0
    for n in self._doing:
      occur[n] = 1
    for n in self._knowing:
      occur[n] = 1
    for n in self._on:
      occur[n] = 1

    return self.latexQuery() + " = " + self._root.toLatex(occur)

  def copy(self) -> "CausalFormula":
    """
    Copy theAST. Note that the causal model is just referenced. The tree is copied.

    :return: the new CausalFormula
    """
    return CausalFormula(self.cm, self.root.copy())

  @property
  def cm(self) -> "CausalModel":
    """
    :return: the causal model
    """
    return self._cm

  @property
  def root(self) -> ASTtree:
    """
    :return: ASTtree root of the CausalFormula tree
    """
    return self._root

  def eval(self) -> gum.Potential:
    """
    Compute the Potential from the CausalFormula over vars using cond as value for others variables

    :param bn: the BN where to infer
    :return:
    """
    return self.root.eval(self.cm.bn())


def _getLabelIdx(bn: gum.BayesNet, varname: str, val: Union[int, str]) -> int:
  """
  Find the index of a label in a discrete variable from a BN.

  If val is an int, we keep is as is. If it is a str, we try to find the correct index in the variable
  :param bn: the BN where to find the variable
  :param varname: the name of the variable
  :param val: the index or the name of the label

  :return: the index of the label
  """
  if not isinstance(val, str):
    return val
  else:
    return bn.variableFromName(varname).index(val)



# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017
#   (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

"""
Basic implementation for dynamic Bayesian Networks in pyAgrum
"""

import matplotlib.pyplot as plt
import numpy as np
import pyAgrum as gum
import pydotplus as dot
from matplotlib.patches import Rectangle

from .notebook import showGraph, getGraph

noTimeCluster = "void"


def _splitName(name):
  """
  By convention, name of dyna:ic variable "X" in dBN may be
    - "X0" for timeslice 0 both in unrolled BN and in 2TBN
    - "Xt" for timeslice t in a 2TBN
    - "X"+str(i) for timelice i with integer i>0 in unrolled BN
    - other naes are not in a timeslice
  @argument name : str (name of the dynamic variable)
  @return static_name,timeslice with timeslice =noTimeCluster,"t" or str(i)
  """
  if name[-1] == "t":
    return name[:-1], "t"
  i = len(name) - 1
  if not name[i].isdigit():
    return name, noTimeCluster

  while name[i].isdigit():
    if i == 0:
      return name, noTimeCluster
    i -= 1

  return name[:i + 1], name[i + 1:]


def _isInFirstTimeSlice(name):
  """
  @return true if there is a 0 at the end of name
  """
  return name[-1] == "0"


def _isInSecondTimeSlice(name):
  """
  @return true if there is a t at the end of name
  """
  return name[-1] == "t"


def _isInNoTimeSlice(name):
  return name[-1] not in ["0", "t"]


def realNameFrom2TBNname(name, ts):
  """
  @return dynamic name from static name and timeslice (no check)
  """
  return "{0}{1}".format(name[:-1], ts) if not _isInNoTimeSlice(name) else name


def getTimeSlicesRange(dbn):
  """
  get the range and (name,radical) of each variables

  :param dbn: a 2TBN or an unrolled BN
  :return: all the timeslice of a dbn

  e.g.
  ['0','t'] for a classic 2TBN
  range(T) for a classic unrolled BN
  """
  timeslices = {}

  for i in dbn.nodes():
    n = dbn.variable(i).name()
    label, ts = _splitName(n)
    if ts in timeslices:
      timeslices[ts].append((n, label))
    else:
      timeslices[ts] = [(n, label)]

  return timeslices


def is2TBN(bn):
  ts = getTimeSlicesRange(bn)
  if not (set(ts.keys()) <= set([noTimeCluster, "0", "t"])):
    return False, "Some variables are not correctly suffixed."

  domainSizes = dict()
  for name, radical in ts['t']:
    domainSizes[radical] = bn.variable(name).domainSize()

  res = ""
  for name, radical in ts['0']:
    if radical in domainSizes:
      if domainSizes[radical] != bn.variable(name).domainSize():
        res = "\n - for variables {}/{}t".format(name, radical)

  if res != "":
    return False, "Domain size mismatch : " + res
  else:
    return True, ""


def _TimeSlicesToDot(dbn):
  """
  Try to correctly represent dBN and 2TBN in dot format
  """
  timeslices = getTimeSlicesRange(dbn)

  g = dot.Dot(graph_type='digraph')
  g.set_rankdir("TD")
  g.set_splines("ortho")

  for k in sorted(timeslices.keys(), key=lambda x: -1 if x == noTimeCluster else 1e8 if x == 't' else int(x)):
    if k != noTimeCluster:
      cluster = dot.Cluster(k, label="Time slice {}".format(k), bgcolor="#DDDDDD", rankdir="TD")
      g.add_subgraph(cluster)
    else:
      cluster = g  # small trick to add in graph variable in no timeslice
    for (n, label) in sorted(timeslices[k]):
      cluster.add_node(dot.Node(n, label=label, style='filled', color='#000000', fillcolor='white'))

  for tail, head in dbn.arcs():
    g.add_edge(dot.Edge(dbn.variable(tail).name(),
                        dbn.variable(head).name(),
                        constraint=False, color="blue"))

  for k in sorted(timeslices.keys(), key=lambda x: -1 if x == noTimeCluster else 1e8 if x == 't' else int(x)):
    if k != noTimeCluster:
      prec = None
      for (n, label) in sorted(timeslices[k]):
        if prec is not None:
          g.add_edge(dot.Edge(prec,
                              n,
                              style="invis"))
        prec = n

  return g


def showTimeSlices(dbn, size="6", format="png"):
  """
  Try to correctly display dBN and 2TBN

  :param dbn: the dynamic BN
  :param size: size of the figue
  :param format: png/svg
  """

  showGraph(_TimeSlicesToDot(dbn), size, format)


def getTimeSlices(dbn, size="6", format="png"):
  """
  Try to correctly represent dBN and 2TBN as an HTML string

  :param dbn: the dynamic BN
  :param size: size of the figue
  :param format: png/svg
  """

  return getGraph(_TimeSlicesToDot(dbn), size, format)


def unroll2TBN(dbn, nbr):
  """
  unroll a 2TBN given the nbr of timeslices

  :param dbn: the dBN
  :param nbr: the number of timeslice

  :return: unrolled BN from a 2TBN and the nbr of timeslices
  """
  ts = getTimeSlicesRange(dbn)
  if not (set([noTimeCluster, "0", "t"]).issuperset(ts.keys())
          and set(["0", "t"]).issubset(ts.keys())):
    raise TypeError("unroll2TBN needs a 2-TimeSlice BN")

  bn = gum.BayesNet()

  # variable creation
  for dbn_id in dbn.nodes():
    name = dbn.variable(dbn_id).name()
    if _isInNoTimeSlice(name):
      bn.add(dbn.variable(dbn_id))
    elif _isInFirstTimeSlice(name):
      bn.add(dbn.variable(dbn_id))  # create a clone of the variable in the new bn
    else:
      for ts in range(1, nbr):
        nid = bn.add(dbn.variable(dbn_id))  # create a clone of the variable in the new bn
        bn.changeVariableName(nid, realNameFrom2TBNname(name, ts))  # create the true name

  # add parents
  # the main pb : to have the same order for parents w.r.t the order in 2TBN
  for dbn_id in dbn.nodes():
    name = dbn.variable(dbn_id).name()
    # right order for parents
    l = dbn.cpt(dbn_id).var_names
    l.pop()
    l.reverse()

    for name_parent in l:
      if not _isInSecondTimeSlice(name):
        if not _isInSecondTimeSlice(name_parent):
          bn.addArc(bn.idFromName(name_parent), bn.idFromName(name))
        else:
          if _isInFirstTimeSlice(name):
            raise TypeError("An arc from timeslice t to timeslice is impossible in dBN")
          else:
            for ts in range(1, nbr):
              new_name_parent = realNameFrom2TBNname(name_parent, ts)  # current TimeSlice
              bn.addArc(bn.idFromName(new_name_parent), bn.idFromName(name))
      else:
        for ts in range(1, nbr):
          if _isInFirstTimeSlice(name_parent):
            new_name_parent = realNameFrom2TBNname(name_parent, ts - 1)  # last TimeSlice
          else:
            new_name_parent = realNameFrom2TBNname(name_parent, ts)  # current TimeSlice
          new_name = realNameFrom2TBNname(name, ts)  # necessary current TimeSlice
          bn.addArc(bn.idFromName(new_name_parent), bn.idFromName(new_name))

  # potential creation
  for dbn_id in dbn.nodes():
    name = dbn.variable(dbn_id).name()
    if not _isInSecondTimeSlice(name):
      bn.cpt(bn.idFromName(name))[:] = dbn.cpt(dbn_id)[:]
    else:
      for ts in range(1, nbr):
        bn.cpt(bn.idFromName(realNameFrom2TBNname(name, ts)))[:] = dbn.cpt(dbn_id)[:]

  return bn


def _plotFollowUnrolled(lovars, dbn, bn, T, evs):
  """
  plot the dynamic evolution of a list of vars with a dBN

  :param lovars: list of variables to follow
  :param dbn: the dbn
  :param bn: the unrolled dbn
  :param T: the time range
  :param evs: observations
  """
  ie = gum.LazyPropagation(bn)
  ie.setEvidence(evs)
  ie.makeInference()

  x = np.arange(T)

  for var in lovars:
    v0 = dbn.variableFromName(var + "0")
    l = []
    for i in range(v0.domainSize()):
      serie = []
      for t in range(T):
        serie.append(ie.posterior(bn.idFromName(var + str(t)))[i])
      l.append(serie)

    fig, ax = plt.subplots()
    plt.ylim(ymax=1, ymin=0)
    ax.xaxis.grid()
    plt.title("Following variable {0}".format(var), fontsize=20)
    plt.xlabel('time')

    stack = ax.stackplot(x, l)

    proxy_rects = [Rectangle((0, 0), 1, 1, fc=pc.get_facecolor()[0]) for pc in stack]
    labels = [v0.label(i) for i in range(v0.domainSize())]
    plt.legend(proxy_rects, labels, loc='center left', bbox_to_anchor=(1, 0.5), ncol=1, fancybox=True, shadow=True)

    plt.show()


def plotFollow(lovars, dbn, T, evs):
  """
  plots modifications of variables in a 2TDN knowing the size of the time window (T) and the evidence on the sequence.

  :param lovars: list of variables to follow
  :param dbn: the dbn
  :param T: the time range
  :param evs: observations
  """
  _plotFollowUnrolled(lovars, dbn, unroll2TBN(dbn, T), T, evs)


if __name__ == '__main__':
  pyAgrum_header("2015-17")

  dbn = gum.BayesNet()
  a0, b0, c0, d0, at, bt, ct, dt = [dbn.add(gum.LabelizedVariable(s, s, 5)) for s in
                                    ["a0", "b0", "c0", "d0", "at", "bt", "ct", "dt"]]

  dbn.addArc(a0, b0)
  dbn.addArc(c0, d0)
  dbn.addArc(at, bt)

  dbn.addArc(a0, bt)
  dbn.addArc(b0, ct)
  dbn.addArc(c0, dt)
  dbn.addArc(d0, at)

  # needed from within spyder, do not understand why
  import time

  gum.initRandom(int(str(time.time()).split('.')[1]))
  # should not be necessary
  dbn.generateCPTs()

  showTimeSlices(dbn)

  # T=5
  # bn=unroll2TBN(dbn,T)
  # showTimeSlices(bn)

  plotFollow(["a", "b", "c", "d"], dbn, T=51, evs={'a15': 2, 'a30': 0, 'c14': 0, 'b40': 0})
